@@ -1,6 +1,9 @@
 import os
+import errno
+from html import escape
 import teampy
 import pandas as pd
+from datetime import date
 from teampy import Questionaire, SolutionDocument, Students, Teams, Solution, Result, Teampy, RATContext, tell
 from colorama import init, Fore, Style
 import click
@@ -14,6 +17,7 @@ import smtplib
 from smtplib import SMTPHeloError, SMTPRecipientsRefused, SMTPNotSupportedError, SMTPAuthenticationError, SMTPSenderRefused, SMTPDataError, SMTPException
 import getpass
 import progressbar
+from latex import build_pdf
 
 def print_teampy():
     print('')
@@ -22,7 +26,7 @@ def print_teampy():
     print('| __/ _ \/ _` | \'_ ` _ \| \'_ \| | | |')
     print('| ||  __/ (_| | | | | | | |_) | |_| |')
     print(' \__\___|\__,_|_| |_| |_| .__/ \__, |')
-    print('                        |_|    |___/ ')
+    print('                        |_|    |___/ {}'.format(teampy.__version__))
     print('')
 
 def copy_figures(rat_directory):
@@ -33,6 +37,16 @@ def copy_figures(rat_directory):
     for filename in ['rat-box.pdf', 'scratch.pdf']:
         file = os.path.join(os.path.dirname(__file__), filename)
         shutil.copy(file, figures)
+        
+def make_sure_path_exists(path):
+    """
+    Ensures that all directories for a given path exist.
+    """
+    try:
+        os.makedirs(path)
+    except OSError as exception:
+        if exception.errno != errno.EEXIST:
+            raise
 
 def rat_create():
     """
@@ -42,10 +56,6 @@ def rat_create():
 
     from prompt_toolkit.shortcuts import input_dialog
 
-    text = input_dialog(
-    title='Input dialog example',
-    text='Please type your name:')
-
     # create directory
     #number = int(prompt('Number of the RAT: ', validator=NumberValidator(), key_bindings=bindings))
     number = prompt('Number of the RAT: ', validator=NumberValidator(), key_bindings=bindings)
@@ -72,16 +82,18 @@ def _load_rat_file(file_input):
     return questionaire
 
 
-def write_latex(latex, file_path):
+def write_latex(latex, file_path, old_latex=False):
     tex_file_name = os.path.splitext(os.path.basename(file_path))[0] + '.tex'
     tex_file_path = os.path.join(os.path.dirname(file_path), tex_file_name)
     # TODO check if file already exists
     with open(tex_file_path, "w", encoding='utf-8') as file:
         file.write(latex)
-    # copy also ratbox figures
-    copy_figures(os.path.dirname(file_path))
+    if old_latex:
+        # copy also ratbox figures
+        copy_figures(os.path.dirname(file_path))
     tell('Write LaTeX into file {}.'.format(tex_file_path))
 
+
 def rat_check(file_input, file_path):
     """
     Read the RAT questionaire, check for consistency, print it as tex document,
@@ -91,7 +103,7 @@ def rat_check(file_input, file_path):
     if questionaire is None:
         return
 
-    # does not detect 4 fake answers, or several true answers, improve parser
+    # TODO does not detect 4 fake answers, or several true answers, improve parser
     tell('The RAT has {} questions.'.format(len(questionaire.questions)))
     for question in questionaire.questions:
         fake = len(question.fake)
@@ -108,9 +120,9 @@ def rat_check(file_input, file_path):
 def parallel_file_path(file_path, alternative_extension):
     head, tail = os.path.split(file_path)
     tail_base = tail.split('.')[0]
-    return os.path.join(os.path.dirname(file_path), '{}.{}'.format(tail_base, alternative_extension))
+    return os.path.join(os.path.dirname(file_path), '{}{}'.format(tail_base, alternative_extension))
 
-def rat_print(file_input, file_path, team_solution):
+def rat_print(file_input, file_path, team_solution, old_latex=False, pdf=True):
     """
     Create a document with RATs for all students and all teams.
     """
@@ -138,11 +150,19 @@ def rat_print(file_input, file_path, team_solution):
     sd.store(solutions_file_path)
     tell('Write solutions into file {}.'.format(solutions_file_path))
 
-    latex = questionaire.write_latex(sd, t.teams, t.students)
-    write_latex(latex, file_path)
-
-    tell('Done! ')
-    print('   As a next step, print the LaTeX document and do the RAT.')
+    latex = questionaire.write_latex(sd, t.teams, t.students, old_latex)
+    if pdf:
+        pdf_file_name = os.path.splitext(os.path.basename(file_path))[0] + '.pdf'
+        tell('Creating PDF...')
+        # empty directory for tex reasons...
+        current_dir = os.path.abspath(os.path.dirname(file_path))
+        pdf = build_pdf(latex, texinputs=[current_dir, ''])
+        pdf.save_to(pdf_file_name)
+        tell('Done!')
+    else:
+        write_latex(latex, file_path, old_latex=old_latex)
+        tell('Done!')
+        print('   As a next step, print the LaTeX document and do the RAT.')
 
 
 def rat_grade(file_input, file_path):
@@ -153,7 +173,7 @@ def rat_grade(file_input, file_path):
     rat = RATContext(teampy)
 
     # read in the questionaire
-    questionaire = _load_rat_file(click.open_file(rat.questionaire_file, encoding='latin-1'))
+    questionaire = _load_rat_file(click.open_file(rat.questionaire_file, encoding='utf-8'))
     if questionaire is None:
         tell('Could not read questions. Aborting.', 'error')
         return
@@ -162,11 +182,20 @@ def rat_grade(file_input, file_path):
     solutions.load(rat.solutions_file, teampy.students, teampy.teams)
 
     result = Result(teampy.students, teampy.teams, questionaire, solutions)
-    result.load_results(file_input)
+    student_results_count, team_results_count = result.load_results(file_input)
 
-    results_file_path = parallel_file_path(file_path, 'xlsx')
+    if (student_results_count + team_results_count) == 0:
+        tell('The results file does not contain any results.', 'warning')
+        return
+
+    results_file_path = parallel_file_path(file_path, '.xlsx')
     result.store_results(results_file_path)
-    result.stats()
+    results_file_path = parallel_file_path(file_path, '.html')
+    result.store_results_html(results_file_path)
+    result.stats(filename=parallel_file_path(file_path, '_stats.html'))
+
+def escape_html_entities(string):
+    return string.encode('ascii', 'xmlcharrefreplace').decode('UTF-8')
 
 def create_message(student_id, row, result, teampy):
     msg = MIMEMultipart()
@@ -174,19 +203,22 @@ def create_message(student_id, row, result, teampy):
     msg['To'] = row['email']
     # TODO here the course code would be nice to have
     msg['Subject'] = 'RAT Feedback'
+    firstname = escape_html_entities(row['firstname'])
+    lastname = escape_html_entities(row['lastname'])
+    student_id = escape_html_entities(student_id)
     html = """\
 <html>
   <head></head>
   <body>
     <p>Hi {}!</p>
-    <p>Here are the results form the latest RAT:</p>
+    <p>Here are the results from the latest RAT:</p>
     <table>
       <tr><td>Name:</td><td>{} {}</td><tr>
       <tr><td>ID:</td><td><code>{}</code></td><tr>
       <tr><td>Team:</td><td><code>{}</code></td><tr>
       <tr><td>RAT:</td><td><code>{}</code></td><tr>
       <tr><td>Date:</td><td><code>{}</code></td><tr>
-""".format(row['firstname'], row['firstname'], row['lastname'], student_id, row['team'], result.name, result.date)
+""".format(firstname, firstname, lastname, student_id, row['team'], result.name, result.date)
     html = html + """\
       <tr><td>Individual Answer:&nbsp;</td><td><code>{}</code></td><tr>
       <tr><td>Correct Solution:</td><td><code>{}</code></td><tr>
@@ -201,18 +233,23 @@ def create_message(student_id, row, result, teampy):
     html = html + """\
       <tr><td>Final Score for this RAT:</td><td><b>{}</b></td><tr>
     </table>
+    <p>{}</p>
     <p>This mail was automated, but you can reply to it.</p>
   </body>
 </html>
-""".format(row['score'])
+""".format(row['score'], row['comment'])
     msg.attach(MIMEText(html, 'html'))
-    return msg    
+    return msg
 
 
-def rat_email(file_input, file_path):
+def rat_email(file_input, file_path, testonly):
     """
     Send the results of a RAT to students via email.
     """
+    if not file_path.endswith('.xlsx'):
+        tell('This command only takes an *.xlsx file as input.', 'error')
+        return
+
     teampy = Teampy()
     
     if teampy.smtp_settings is None:
@@ -222,7 +259,7 @@ def rat_email(file_input, file_path):
     # TODO give RAT context the directory relative to the file_input
     rat = RATContext(teampy)
     # read in the questionaire
-    questionaire = _load_rat_file(click.open_file(rat.questionaire_file, encoding='latin-1'))
+    questionaire = _load_rat_file(click.open_file(rat.questionaire_file, encoding='utf-8'))
     if questionaire is None:
         tell('Could not read questions. Aborting.', 'error')
         return
@@ -231,9 +268,9 @@ def rat_email(file_input, file_path):
     solutions.load(rat.solutions_file, teampy.students, teampy.teams)
 
     # read in the corresponding result file
-    results_file_path = parallel_file_path(file_path, 'txt')
+    results_file_path = parallel_file_path(file_path, '.txt')
     result = Result(teampy.students, teampy.teams, questionaire, solutions)
-    result.load_results(click.open_file(results_file_path, encoding='latin-1'))
+    result.load_results(click.open_file(results_file_path, encoding='utf-8'))
 
     # read in the graded file
     df = pd.read_excel(file_path, dtype={'id': str, 'team': str, 'email':str, 'comment':str, 'feedback':str})
@@ -249,24 +286,38 @@ def rat_email(file_input, file_path):
         # TODO check if we need to send email to this student
         if row['feedback'] != 'sent':
             messages[student_id] = create_message(student_id, row, result, teampy)
+
+    # store the messages
+    make_sure_path_exists('emails')
+    for student_id, message in messages.items():
+        html_file_path = os.path.join(os.path.dirname(file_path), 'emails/{}.html'.format(student_id))
+        with open(html_file_path, "w") as html_file:
+            #html_file.write(message.as_string())
+            html_file.write(message.get_payload()[0].get_payload())
+
     if len(messages)==0:
         tell('There is nothing to send.')
         return
+    elif testonly:
+        tell('Messages written to email folder, but nothing will be sent due to option --testonly.')
+        return
     else:
-        tell('Will send messages to {} students.'.format(len(messages)))
+        tell('Will send messages to {} students. (Press CTRL+C to quit and not send anything. Restart with option --testonly to only store a preview of the emails.)'.format(len(messages)))
 
     # connect to SMTP server
     print('\n')
     password = getpass.getpass(prompt='Password for the user {} on server {}: '.format(teampy.smtp_settings['from'], teampy.smtp_settings['smtp']))
 
     #create server
+    print('Create client...')
     server = smtplib.SMTP('{}: {}'.format(teampy.smtp_settings['smtp'], teampy.smtp_settings['port']))
     statuses = {}
     try:
+        print('Connecting to server...')
         server.starttls()
         # Login Credentials for sending the mail
         server.login(teampy.smtp_settings['from'], password)
-
+        print('Logged in...')
         # send email
         bar = progressbar.ProgressBar(max_value=len(messages), widgets=[progressbar.Percentage(), progressbar.Bar()])
         for student_id, message in messages.items():
@@ -324,7 +375,7 @@ def new():
     """
     print_teampy()
     click.echo('Create a new RAT.')
-    # store as ISO-Latin-1
+    # store as UTF-8
     # ask for a title
     # ask for number of questions
     # ask for answer alternatives
@@ -336,44 +387,107 @@ def new():
     # a complete example document?
 
 @rat.command()
-#@click.argument('file', type=click.File('r'))
 @click.argument('file', type=click.Path(exists=True))
 def check(file):
     """
     Check a RAT file for consistency before printing.
     """
     print_teampy()
-    rat_check(click.open_file(file, encoding='latin-1'), file)
+    rat_check(click.open_file(file, encoding='utf-8'), file)
+
+@rat.command()
+@click.argument('file', type=click.Path(exists=True))
+def trial(file):
+    """
+    Print the RAT for a trial run, for instance with a colleague.
+    Question sequence is original, but answers are shuffled.
+    """
+    print_teampy()
+    file_path = file
+    file = click.open_file(file, encoding='utf-8')
+    questionaire = _load_rat_file(file)
+    if questionaire is None:
+        return
+    latex = questionaire.write_trial_latex()
+    write_latex(latex, file_path)
 
 @rat.command(name='print')
 #@click.argument('file', type=click.File('r'))
-@click.argument('file', type=click.Path(exists=True))
+@click.argument('file', type=click.Path(exists=True), required=False)
+@click.option('--nopdf', default=False, is_flag=True, help='Do not create PDF, just latex source.')
 @click.option('--teamsolution', prompt='Team solution', help='Code of the team scratch card or team solution.')
-def print_(file, teamsolution):
+def print_(file, teamsolution, nopdf):
     """
     Print a RAT before class.
     """
     # TODO if no valid team solution code is shown, add prompt that also shows which scratch cards are available
     print_teampy()
-    rat_print(click.open_file(file, encoding='latin-1'), file, teamsolution)
+    pdf = not nopdf
+    if file is None:
+        file = os.path.join(os.getcwd(), 'questions.txt')
+        if os.path.exists(file):
+            if click.confirm('Found file {}. Do you want to print it?'.format(file)):
+                rat_print(click.open_file(file, encoding='utf-8'), file, teamsolution, pdf=pdf)
+    else:
+        rat_print(click.open_file(file, encoding='utf-8'), file, teamsolution, pdf=pdf)
+
+
+def rat_setup_results_file(result_file):
+    with open(result_file, "w", encoding='utf-8') as file:
+        file.write(('---\n'
+         'name: RAT\n'
+         'date: {}\n'
+         '---\n'.format(date.today().strftime("%Y-%m-%d"))))
+
 
 @rat.command()
-@click.argument('file', type=click.Path(exists=True))
+@click.argument('file', type=click.Path(exists=True), required=False)
 def grade(file):
     """
     Evaluate a RAT during class.
     """
     print_teampy()
-    rat_grade(click.open_file(file, encoding='latin-1'), file)
+    if file is None:
+        file = os.path.join(os.getcwd(), 'results.txt')
+        if os.path.exists(file):
+            if click.confirm('Found file {}. Do you want to grade it?'.format(file)):
+                rat_grade(click.open_file(file, encoding='utf-8'), file)
+        else:
+            if click.confirm('No results file found or specified. Should we create an empty one?'.format(file)):
+                rat_setup_results_file(file) 
+    else:
+        rat_grade(click.open_file(file, encoding='utf-8'), file)
+
 
 @rat.command()
 @click.argument('file', type=click.Path(exists=True))
-def email(file):
+@click.option('--testonly', default=False, is_flag=True, help='Only store HTML messages for testing, do not actually send.')
+def email(file, testonly):
     """
     Send feedback to students via email.
     """
     print_teampy()
-    rat_email(click.open_file(file, encoding='latin-1'), file)
+    rat_email(click.open_file(file, encoding='utf-8'), file, testonly)
+
+
+@rat.command()
+@click.argument('file', type=click.Path(exists=True))
+@click.option('--format', type=click.Choice(['blackboard'], case_sensitive=False))
+def export(file, format):
+    """
+    Export the questions to another format.
+    """
+    t = Teampy()
+    file_input = click.open_file(file, encoding='utf-8')
+    questionaire = _load_rat_file(file_input)
+    if questionaire is None:
+        return
+    if format == 'blackboard':
+        text = questionaire.write_blackboard()
+        export_file_path = os.path.join(os.path.dirname(file), 'blackboard.txt')
+        with open(export_file_path, "w", encoding='utf-8') as file:
+            file.write(text)
+
 
 if __name__ == "__main__":
     rat()
